MutView

Struct MutView 

pub struct MutView<'a, B: BufferMut + ?Sized> { /* private fields */ }
A writable receive buffer that scatters incoming bytes into a base buffer according to a UserDatatype layout. Mirrors rsmpi’s MutView.

Implementations

Source

impl<'a, B: BufferMut + ?Sized> MutView<'a, B>

pub fn with_count( buf: &'a mut B, datatype: UserDatatype, count: Count, ) -> MutView<'a, B>

Create a mutable view of count elements of datatype over buf.