wait_any

Function wait_any 

pub fn wait_any<'a, D: ?Sized, S: Scope<'a>>(
    requests: &mut Vec<Request<'a, D, S>>,
) -> Option<(usize, Status)>
Wait for any one of the requests to complete, returning its index and status and removing it from the vector (MPI_Waitany). Returns None if the vector is empty.