Scope

Trait Scope 

pub unsafe trait Scope<'a> { }
A scope that bounds the lifetime of the buffers borrowed by requests. Safe scopes are StaticScope ('static) and the LocalScope handed to the closure passed to scope.

Safety

Implementors guarantee that any buffer associated with a request created in this scope outlives the scope.

Dyn Compatibility

This trait is dyn compatible.

In older versions of Rust, dyn compatibility was called "object safety".

Implementors

Source

impl Scope<'static> for StaticScope

Source

impl<'a> Scope<'a> for &LocalScope<'a>