pub unsafe trait Partitioned {
// Required methods
fn as_datatype(&self) -> DatatypeRef;
fn counts(&self) -> &[Count] ⓘ;
fn displs(&self) -> &[Count] ⓘ;
}Expand description
A buffer whose elements are partitioned into per-rank blocks of possibly differing sizes, described by counts and displacements. Used by the varying-count (“v”) collectives.
Safety
The counts and displacements must describe blocks that lie within the underlying buffer.
Required Methods
Sourcefn as_datatype(&self) -> DatatypeRef
fn as_datatype(&self) -> DatatypeRef
The datatype of one element.
Dyn Compatibility
This trait is dyn compatible.
In older versions of Rust, dyn compatibility was called "object safety".